@@ -144,7 +144,7 @@ int flag_detailed_statistics;
 int flag_all_virtual;
 
 /* Zero means that `this' is a *const.  This gives nice behavior in the
-   2.0 world.  Nonzero gives 1.2-compatible behavior.  */
+   2.0 world.  1 gives 1.2-compatible behavior.  2 gives Spring behavior.   */
 
 int flag_this_is_variable;
 
@@ -156,7 +156,9 @@ int flag_memoize_lookups; int flag_save_memoized_contexts;
 
 int flag_dynamic;
 
-/* 2 means write out only specific virtual function tables
+/* 3 means write out only virtuals function tables `defined'
+   in this implementation file.
+   2 means write out only specific virtual function tables
    and give them (C) public visibility.
    1 means write out virtual function tables and give them
    (C) public visibility.
@@ -173,7 +175,12 @@ int flag_elide_constructors;
 /* Nonzero means if the type has methods, only output debugging
    information if methods are actually written to the asm file.  */
 
-int flag_minimal_debug;
+int flag_minimal_debug = 0;
+
+/* Same, but for inline functions: nonzero means write out debug info
+   for inlines.  Zero means do not.  */
+
+int flag_inline_debug;
 
 /* Nonzero means recognize and handle exception handling constructs.
    2 means handle exceptions the way Spring wants them handled.  */
@@ -194,6 +201,10 @@ int flag_no_inline = 0;
    0 means enums can convert to ints, but not vice-versa.  */
 int flag_int_enum_equivalence;
 
+/* Controls whether compiler is operating under LUCID's Cadillac
+   system.  1 means yes, 0 means no.  */
+int flag_cadillac;
+
 /* Table of language-dependent -f options.
    STRING is the option name.  VARIABLE is the address of the variable.
    ON_VALUE is the value to store in VARIABLE
@@ -216,6 +227,7 @@ static struct { char *string; int *variable; int on_value;} lang_f_options[] =
   {"memoize-lookups", &flag_memoize_lookups, 1},
   {"elide-constructors", &flag_elide_constructors, 1},
   {"minimal-debug", &flag_minimal_debug, 1},
+  {"inline-debug", &flag_inline_debug, 0},
   {"handle-exceptions", &flag_handle_exceptions, 1},
   {"spring-exceptions", &flag_handle_exceptions, 2},
   {"default-inline", &flag_default_inline, 1},
@@ -271,6 +283,16 @@ lang_decode_option (p)
 	  flag_all_virtual = 2;
 	  found = 1;
 	}
+      else if (! strncmp (p, "cadillac", 8))
+	{
+	  flag_cadillac = atoi (p+9);
+	  found = 1;
+	}
+      else if (! strncmp (p, "no-cadillac", 11))
+	{
+	  flag_cadillac = 0;
+	  found = 1;
+	}
       else for (j = 0;
 		!found && j < sizeof (lang_f_options) / sizeof (lang_f_options[0]);
 		j++)
@@ -324,6 +346,7 @@ lang_decode_option (p)
 #if 0
 	  warn_enum_clash = 1;
 #endif
+	  warn_inline = 1;
 	}
 
       else if (!strcmp (p, "overloaded-virtual"))
@@ -696,6 +719,9 @@ grokfield (declarator, declspecs, raises, init, asmspec_tree)
       return void_type_node;
     }
 
+  if (flag_cadillac)
+    cadillac_start_decl (value);
+
   if (asmspec_tree)
     asmspec = TREE_STRING_POINTER (asmspec_tree);
 
@@ -707,8 +733,8 @@ grokfield (declarator, declspecs, raises, init, asmspec_tree)
 	  init = NULL_TREE;
 	}
       else if (pedantic
-	  && ! TREE_STATIC (value)
-	  && ! TREE_READONLY (value))
+	       && ! TREE_STATIC (value)
+	       && ! TREE_READONLY (value))
 	{
 	  error ("fields cannot have initializers");
 	  init = error_mark_node;
@@ -718,7 +744,7 @@ grokfield (declarator, declspecs, raises, init, asmspec_tree)
 	  /* We allow initializers to become parameters to base initializers.  */
 	  if (TREE_CODE (init) == CONST_DECL)
 	    init = DECL_INITIAL (init);
-	  else if (TREE_READONLY (init) && TREE_CODE (init) == VAR_DECL)
+	  else if (TREE_READONLY_DECL_P (init))
 	    init = decl_constant_value (init);
 	  else if (TREE_CODE (init) == CONSTRUCTOR)
 	    init = digest_init (TREE_TYPE (value), init, 0);
@@ -770,6 +796,22 @@ grokfield (declarator, declspecs, raises, init, asmspec_tree)
 	      asmspec_tree = build_string (IDENTIFIER_LENGTH (name), asmspec);
 	    }
 	  pending_statics = perm_tree_cons (NULL_TREE, value, pending_statics);
+
+	  /* Static consts need not be initialized in the class definition.  */
+	  if (init != NULL_TREE && TYPE_NEEDS_CONSTRUCTING (TREE_TYPE (value)))
+	    {
+	      static int explanation = 0;
+
+	      error ("initializer invalid for static member with constructor");
+	      if (explanation++ == 0)
+		error ("(you really want to initialize it separately)");
+	      init = 0;
+	    }
+	  /* Force the user to know when an uninitialized static
+	     member is being used.  */
+	  if (TREE_READONLY (value)
+	      && TYPE_NEEDS_CONSTRUCTING (TREE_TYPE (value)))
+	    TREE_EXTERNAL (value) = 1;
 	}
       DECL_INITIAL (value) = init;
       DECL_IN_AGGR_P (value) = 1;
@@ -826,12 +868,22 @@ grokbitfield (declarator, declspecs, width)
   if (TREE_CODE (value) == VOID_TYPE)
     return void_type_node;
 
+  if (TREE_CODE (value) == TYPE_DECL)
+    {
+      error_with_decl (value, "cannot declare `%s' to be a bitfield type");
+      return NULL_TREE;
+    }
+
   if (DECL_IN_AGGR_P (value))
     {
       error_with_decl (value, "`%s' is already defined in aggregate scope");
       return void_type_node;
     }
 
+#ifdef FIELD_XREF
+  FIELD_xref_member(current_class_name,value);
+#endif
+
   if (TREE_STATIC (value))
     {
       error_with_decl (value, "static member `%s' cannot be a bitfield");
@@ -843,7 +895,7 @@ grokbitfield (declarator, declspecs, width)
       /* detect invalid field size.  */
       if (TREE_CODE (width) == CONST_DECL)
 	width = DECL_INITIAL (width);
-      else if (TREE_READONLY (width) && TREE_CODE (width) == VAR_DECL)
+      else if (TREE_READONLY_DECL_P (width))
 	width = decl_constant_value (width);
 
       if (TREE_CODE (width) != INTEGER_CST)
@@ -967,7 +1019,13 @@ groktypefield (declspecs, parmlist)
 	    }
 	}
       else if (TREE_CODE (id) == RECORD_TYPE)
-	error ("identifier for aggregate type conversion omitted");
+	{
+	  type_id = TYPE_NAME (id);
+	  if (TREE_CODE (type_id) == TYPE_DECL)
+	    type_id = DECL_NAME (type_id);
+	  if (type_id == NULL_TREE)
+	    error ("identifier for aggregate type conversion omitted");
+	}
       else
 	assert (0);
       spec = TREE_CHAIN (spec);
@@ -1070,15 +1128,13 @@ grokoptypename (decl, call_required)
 				  TREE_PURPOSE (TREE_VALUE (parms)),
 				  TYPENAME, 0, NULL_TREE);
 	    if (parms)
-	      if (parms != void_list_node)
-		{
+	      {
+		if (TREE_VALUE (parms) != void_type_node)
 		  error ("operator <typename> requires empty parameter list");
+		else
+		  /* Canonicalize parameter lists.  */
 		  TREE_OPERAND (tmp, 1) = void_list_node;
-		}
-	      else
-		/* If user specifies `void' explicitly, avoid having
-		   two voids, since we put one invisible one on.  */
-		TREE_CHAIN (parms) = NULL_TREE;
+	      }
 
 	    last = grokdeclarator (TREE_OPERAND (decl, 0),
 				   TREE_TYPE (decl),
@@ -1169,7 +1225,7 @@ grokopexpr (declp, ctype, type, report_ambiguous, is_decl)
 	  if (TYPE_ARG_TYPES (type)
 	      && TREE_CHAIN (TYPE_ARG_TYPES (type))
 	      && TREE_CHAIN (TYPE_ARG_TYPES (type)) != void_list_node)
-	    return get_identifier ("__user_new");
+	    return get_identifier (OPERATOR_NEW_FORMAT);
 	  return get_identifier ("__builtin_new");
 	}
       else if (ctype == 0 && TREE_CODE (TREE_OPERAND (declarator, 0)) == DELETE_EXPR)
@@ -1345,7 +1401,13 @@ grok_function_init (decl, init)
 void
 setup_vtbl_ptr ()
 {
-  if (! flag_this_is_variable
+  extern struct rtx_def *base_init_insns;
+
+  if (base_init_insns == 0
+      && DECL_CONSTRUCTOR_P (current_function_decl))
+    emit_base_init (current_class_type, 0);
+
+  if ((flag_this_is_variable & 1) == 0
       && optimize
       && current_class_type
       && CLASSTYPE_VSIZE (current_class_type)
@@ -1355,7 +1417,9 @@ setup_vtbl_ptr ()
       current_vtable_decl = CLASSTYPE_VTBL_PTR (current_class_type);
       DECL_RTL (current_vtable_decl) = 0;
       DECL_INITIAL (current_vtable_decl) = error_mark_node;
-      finish_decl (current_vtable_decl, vfield, 0);
+      /* Have to cast the initializer, since it may have come from a
+	 more base class then we ascribe CURRENT_VTABLE_DECL to be.  */
+      finish_decl (current_vtable_decl, convert_force (TREE_TYPE (current_vtable_decl), vfield), 0);
       current_vtable_decl = build_indirect_ref (current_vtable_decl, 0);
     }
   else
@@ -1504,6 +1568,9 @@ finish_anon_union (anon_union_decl)
   if (static_p)
     make_decl_rtl (decl, 0, global_bindings_p ());
   expand_anon_union_decl (decl, NULL_TREE, elems);
+
+  if (flag_cadillac)
+    cadillac_finish_anon_union (decl);
 }
 
 /* Finish and output a table which is generated by the compiler.
@@ -1572,43 +1639,152 @@ coerce_delete_type (ctype, type)
      tree type;
 {
   int e1 = 0, e2 = 0, e3 = 0;
+  tree arg_types = TYPE_ARG_TYPES (type);
 
   if (TREE_CODE (type) == METHOD_TYPE)
-    type = build_function_type (TREE_TYPE (type), TREE_CHAIN (TYPE_ARG_TYPES (type)));
+    {
+      type = build_function_type (TREE_TYPE (type), TREE_CHAIN (arg_types));
+      arg_types = TREE_CHAIN (arg_types);
+    }
   if (TREE_TYPE (type) != void_type_node)
     e1 = 1, error ("`operator delete' must return type `void'");
-  if (TYPE_ARG_TYPES (type) == NULL_TREE
-      || TREE_VALUE (TYPE_ARG_TYPES (type)) != ptr_type_node)
+  if (arg_types == NULL_TREE
+      || TREE_VALUE (arg_types) != ptr_type_node)
     e2 = 1, error ("`operator delete' takes type `void *' as first parameter");
 
-  if (TYPE_ARG_TYPES (type)
-      && TREE_CHAIN (TYPE_ARG_TYPES (type))
-      && TREE_CHAIN (TYPE_ARG_TYPES (type)) != void_list_node)
+  if (arg_types
+      && TREE_CHAIN (arg_types)
+      && TREE_CHAIN (arg_types) != void_list_node)
     {
       /* Again, technically this argument must be `size_t', but again
 	 we may not know what that is.  */
-      tree t2 = TREE_VALUE (TREE_CHAIN (TYPE_ARG_TYPES (type)));
+      tree t2 = TREE_VALUE (TREE_CHAIN (arg_types));
       if (TREE_CODE (t2) != INTEGER_TYPE
 	  || TYPE_PRECISION (t2) != TYPE_PRECISION (sizetype))
 	e3 = 1, error ("second argument to `operator delete' must be of type `size_t'");
-      else if (TREE_CHAIN (TREE_CHAIN (TYPE_ARG_TYPES (type))) != void_list_node)
+      else if (TREE_CHAIN (TREE_CHAIN (arg_types)) != void_list_node)
 	{
 	  e3 = 1;
-	  if (TREE_CHAIN (TREE_CHAIN (TYPE_ARG_TYPES (type))))
+	  if (TREE_CHAIN (TREE_CHAIN (arg_types)))
 	    error ("too many arguments in declaration of `operator delete'");
 	  else
 	    error ("`...' invalid in specification of `operator delete'");
 	}
     }
   if (e3)
-    type = build_function_type (void_type_node, tree_cons (NULL_TREE, ptr_type_node, build_tree_list (NULL_TREE, sizetype)));
-  else if (e2)
-    type = build_function_type (void_type_node, tree_cons (NULL_TREE, ptr_type_node, TREE_CHAIN (TYPE_ARG_TYPES (type))));
-  else if (e1)
-    type = build_function_type (void_type_node, TYPE_ARG_TYPES (type));
+    arg_types = tree_cons (NULL_TREE, ptr_type_node, build_tree_list (NULL_TREE, sizetype));
+  else if (e3 |= e2)
+    {
+      if (arg_types == NULL_TREE)
+	arg_types = void_list_node;
+      else
+	arg_types = tree_cons (NULL_TREE, ptr_type_node, TREE_CHAIN (arg_types));
+    }
+  else e3 |= e1;
+
+  if (e3)
+    type = build_function_type (void_type_node, arg_types);
+
   return type;
 }
 
+static void
+write_vtable_entries (decl)
+     tree decl;
+{
+  tree entries;
+
+  for (entries = TREE_CHAIN (CONSTRUCTOR_ELTS (DECL_INITIAL (decl)));
+       entries; entries = TREE_CHAIN (entries))
+    {
+      tree fnaddr = FNADDR_FROM_VTABLE_ENTRY (TREE_VALUE (entries));
+      tree fn = TREE_OPERAND (fnaddr, 0);
+      if (! TREE_ASM_WRITTEN (fn)
+	  && DECL_SAVED_INSNS (fn))
+	{
+	  if (TREE_PUBLIC (DECL_CONTEXT (fn)))
+	    TREE_PUBLIC (fn) = 1;
+	  TREE_ADDRESSABLE (fn) = 1;
+	  output_inline_function (fn);
+	}
+    }
+}
+
+static void
+finish_vtable_typedecl (prev, vars)
+     tree prev, vars;
+{
+  tree decl = CLASS_ASSOC_VTABLE (TREE_TYPE (vars));
+
+  /* If we are controlled by `+e2', obey.  */
+  if (write_virtuals == 2)
+    {
+      tree assoc = value_member (DECL_NAME (vars), pending_vtables);
+      if (assoc)
+	TREE_PURPOSE (assoc) = void_type_node;
+      else
+	decl = NULL_TREE;
+    }
+  /* If this type has inline virtual functions, then
+     write those functions out now.  */
+  if (decl && (TREE_PUBLIC (decl)
+	       || (! TREE_EXTERNAL (decl) && TREE_USED (decl))))
+    write_vtable_entries (decl);
+}
+
+static void
+finish_vtable_vardecl (prev, vars)
+     tree prev, vars;
+{
+  if (write_virtuals < 0)
+    ;
+  else if (write_virtuals == 0
+	   ? TREE_USED (vars)
+	   : (TREE_PUBLIC (vars)
+	      || (! TREE_EXTERNAL (vars) && TREE_USED (vars))))
+    {
+      extern tree the_null_vtable_entry;
+
+      /* Stuff this virtual function table's size into
+	 `pfn' slot of `the_null_vtable_entry'.  */
+      tree nelts = array_type_nelts (TREE_TYPE (vars));
+      tree *ppfn = &FNADDR_FROM_VTABLE_ENTRY (the_null_vtable_entry);
+      *ppfn = nelts;
+      assert (TREE_VALUE (TREE_CHAIN (TREE_CHAIN (CONSTRUCTOR_ELTS (TREE_VALUE (CONSTRUCTOR_ELTS (DECL_INITIAL (vars))))))) == nelts);
+      /* Write it out.  */
+      write_vtable_entries (vars);
+      if (TREE_TYPE (DECL_INITIAL (vars)) == 0)
+	store_init_value (vars, DECL_INITIAL (vars));
+      rest_of_decl_compilation (vars, 0, 1, 1);
+    }
+  /* We know that PREV must be non-zero here.  */
+  TREE_CHAIN (prev) = TREE_CHAIN (vars);
+}
+
+void
+walk_vtables (typedecl_fn, vardecl_fn)
+     register void (*typedecl_fn)();
+     register void (*vardecl_fn)();
+{
+  tree prev, vars;
+
+  for (prev = 0, vars = getdecls (); vars; vars = TREE_CHAIN (vars))
+    {
+      if (TREE_CODE (vars) == TYPE_DECL
+	  && TYPE_LANG_SPECIFIC (TREE_TYPE (vars))
+	  && CLASSTYPE_VSIZE (TREE_TYPE (vars)))
+	{
+	  if (typedecl_fn) (*typedecl_fn) (prev, vars);
+	}
+      else if (TREE_CODE (vars) == VAR_DECL && DECL_VIRTUAL_P (vars))
+	{
+	  if (vardecl_fn) (*vardecl_fn) (prev, vars);
+	}
+      else
+	prev = vars;
+    }
+}
+
 extern int parse_time, varconst_time;
 
 #define TIMEVAR(VAR, BODY)    \
@@ -1746,16 +1922,11 @@ finish_file ()
 	  else
 	    {
 	      mark_addressable (decl);
-	      temp = build (ADDR_EXPR, TYPE_POINTER_TO (type), decl);
+	      temp = build1 (ADDR_EXPR, TYPE_POINTER_TO (type), decl);
 	    }
-	  temp = build_delete (TREE_TYPE (temp), temp,
-			       integer_zero_node, LOOKUP_NORMAL, 0);
+	  temp = build_delete (TREE_TYPE (temp), temp, integer_two_node, LOOKUP_NORMAL, 0);
 	  expand_expr_stmt (temp);
 
-	  if (TYPE_LANG_SPECIFIC (type)
-	      && TYPE_USES_VIRTUAL_BASECLASSES (type))
-	    expand_expr_stmt (build_vbase_delete (type, decl));
-
 	  if (TREE_STATIC (vars))
 	    expand_end_cond ();
 	}
@@ -1768,7 +1939,7 @@ finish_file ()
 
   finish_function (lineno, 0);
 
-#if defined(DBX_DEBUGGING_INFO) && !defined(FASCIST_ASSEMBLER)
+#if defined(DBX_DEBUGGING_INFO) && !defined(FASCIST_ASSEMBLER) && !defined(SDB_DEBUGGING_INFO)
   /* Now tell GNU LD that this is part of the static destructor set.  */
   {
     extern struct _iob *asm_out_file;
@@ -1907,7 +2078,7 @@ finish_file ()
       pop_momentary ();
 
       finish_function (lineno, 0);
-#if defined(DBX_DEBUGGING_INFO) && !defined(FASCIST_ASSEMBLER)
+#if defined(DBX_DEBUGGING_INFO) && !defined(FASCIST_ASSEMBLER) && !defined(SDB_DEBUGGING_INFO)
       /* Now tell GNU LD that this is part of the static constructor set.  */
       {
 	extern struct _iob *asm_out_file;
@@ -2118,91 +2289,30 @@ finish_file ()
 
   /* Now delete from the chain of variables all virtual function tables.
      We output them all ourselves, because each will be treated specially.  */
-  {
-    tree prev;
-    /* Make last thing in global scope not be a virtual function table.  */
-    prev = build_decl (VAR_DECL, get_identifier (" @%$#@!"), integer_type_node);
-    TREE_EXTERNAL (prev) = 1;
-    pushdecl (prev);
-    /* Make this dummy variable look used.  */
-    TREE_ASM_WRITTEN (prev) = 1;
-    DECL_INITIAL (prev) = error_mark_node;
 
-    for (prev = 0, vars = getdecls ();
-	 vars;
-	 vars = TREE_CHAIN (vars))
-      {
-	if (TREE_CODE (vars) == TYPE_DECL
-	    && TYPE_LANG_SPECIFIC (TREE_TYPE (vars))
-	    && CLASSTYPE_VSIZE (TREE_TYPE (vars)))
-	  {
-	    tree decl = CLASS_ASSOC_VTABLE (TREE_TYPE (vars));
+  /* Make last thing in global scope not be a virtual function table.  */
+  vars = build_decl (VAR_DECL, get_identifier (" @%$#@!"), integer_type_node);
+  TREE_EXTERNAL (vars) = 1;
+  pushdecl (vars);
+  /* Make this dummy variable look used.  */
+  TREE_ASM_WRITTEN (vars) = 1;
+  DECL_INITIAL (vars) = error_mark_node;
 
-	    /* If we are controlled by `+e2', obey.  */
-	    if (write_virtuals == 2)
-	      {
-		tree assoc = value_member (DECL_NAME (vars), pending_vtables);
-		if (assoc)
-		  TREE_PURPOSE (assoc) = void_type_node;
-		else
-		  decl = NULL_TREE;
-	      }
-	    /* If this type has inline virtual functions, then
-	       write those functions out now.  */
-	    if (decl && ! TREE_EXTERNAL (decl) && TREE_USED (decl))
-	      {
-		tree entries;
+  walk_vtables (finish_vtable_typedecl, finish_vtable_vardecl);
 
-		for (entries = TREE_CHAIN (CONSTRUCTOR_ELTS (DECL_INITIAL (decl)));
-		     entries; entries = TREE_CHAIN (entries))
-		  {
-		    tree fnaddr = FNADDR_FROM_VTABLE_ENTRY (TREE_VALUE (entries));
-		    tree fn = TREE_OPERAND (fnaddr, 0);
-		    if (! TREE_ASM_WRITTEN (fn)
-			&& DECL_PENDING_INLINE_INFO (fn)
-			&& TREE_ADDRESSABLE (fn))
-		      output_inline_function (fn);
-		  }
-	      }
-	  }
-	if (TREE_CODE (vars) == VAR_DECL && DECL_VIRTUAL_P (vars))
-	  {
-	    if ((write_virtuals == 2
-		 && value_member (DECL_NAME (TYPE_NAME (DECL_VPARENT (vars))),
-				  pending_vtables))
-		|| write_virtuals == 1
-		|| (write_virtuals == 0 && TREE_USED (vars)))
-	      {
-		extern tree the_null_vtable_entry;
+  if (write_virtuals == 2)
+    {
+      /* Now complain about an virtual function tables promised
+	 but not delivered.  */
+      while (pending_vtables)
+	{
+	  if (TREE_PURPOSE (pending_vtables) == NULL_TREE)
+	    error ("virtual function table for `%s' not defined",
+		   IDENTIFIER_POINTER (TREE_VALUE (pending_vtables)));
+	  pending_vtables = TREE_CHAIN (pending_vtables);
+	}
+    }
 
-		/* Stuff this virtual function table's size into
-		   `pfn' slot of `the_null_vtable_entry'.  */
-		tree nelts = array_type_nelts (TREE_TYPE (vars));
-		tree *ppfn = &FNADDR_FROM_VTABLE_ENTRY (the_null_vtable_entry);
-		*ppfn = nelts;
-		assert (TREE_VALUE (TREE_CHAIN (TREE_CHAIN (CONSTRUCTOR_ELTS (TREE_VALUE (CONSTRUCTOR_ELTS (DECL_INITIAL (vars))))))) == nelts);
-		/* Write it out.  */
-		rest_of_decl_compilation (vars, 0, 1, 1);
-	      }
-	    /* We know that PREV must be non-zero here.  */
-	    TREE_CHAIN (prev) = TREE_CHAIN (vars);
-	  }
-	else
-	  prev = vars;
-      }
-    if (write_virtuals == 2)
-      {
-	/* Now complain about an virtual function tables promised
-	   but not delivered.  */
-	while (pending_vtables)
-	  {
-	    if (TREE_PURPOSE (pending_vtables) == NULL_TREE)
-	      error ("virtual function table for `%s' not defined",
-		     IDENTIFIER_POINTER (TREE_VALUE (pending_vtables)));
-	    pending_vtables = TREE_CHAIN (pending_vtables);
-	  }
-      }
-  }
   permanent_allocation ();
   this_time = gettime ();
   parse_time -= this_time - start_time;
