@@ -456,7 +456,7 @@ digest_init (type, init, tail)
 	warning ("ANSI C forbids initializing array from array expression");
       if (TREE_CODE (init) == CONST_DECL)
 	init = DECL_INITIAL (init);
-      else if (TREE_READONLY (init) && TREE_CODE (init) == VAR_DECL)
+      else if (TREE_READONLY_DECL_P (init))
 	init = decl_constant_value (init);
       return init;
     }
@@ -471,7 +471,7 @@ digest_init (type, init, tail)
 	warning ("single-expression nonscalar initializer has braces");
       if (TREE_CODE (element) == CONST_DECL)
 	element = DECL_INITIAL (element);
-      else if (TREE_READONLY (element) && TREE_CODE (element) == VAR_DECL)
+      else if (TREE_READONLY_DECL_P (element))
 	element = decl_constant_value (element);
       return element;
     }
@@ -1097,7 +1097,7 @@ build_functional_cast (exp, parms)
     {
       name = TYPE_NAME (type);
       if (TREE_CODE (name) == TYPE_DECL)
-	name = DECL_NAME (type);
+	name = DECL_NAME (name);
     }
 
   if (TYPE_SIZE (type) == NULL_TREE)
@@ -1364,6 +1364,9 @@ merge_component_comparisons (code, op0, op1)
 	/* Failure.  */
 	return 0;
 
+      if (align < bitsize)
+	return 0;
+
       /* Figure out the mode we will use for this access.  */
       for (trymode = QImode;
 	   trymode && GET_MODE_SIZE (trymode) * BITS_PER_UNIT < bitsize;
