@@ -48,6 +48,7 @@ the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.  */
 #undef FLOAT
 #include <sys/times.h>
 #include <time.h>   /* Correct for hpux at least.  Is it good on other USG?  */
+#undef FFS  /* Some systems define this in param.h.  */
 #else
 #ifndef VMS
 #include <sys/time.h>
@@ -71,6 +72,8 @@ the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.  */
 int been_here_before = 0;	/* Is this an unexec'd program? */
 int just_done_unexec = 0;	/* Is this the process that *was* unexec'd?  */
 char *dump_source_name;		/* Keep argv[0] for dumping. */
+
+/* @@ Should there be a filename length limit?  */
 char previous_asm_file_name[128]; /* asm_file_name from the first run. */
 
 extern int yydebug;
@@ -85,6 +88,7 @@ extern void init_decl_processing ();
 extern void init_tree ();
 extern void init_rtl ();
 extern void init_optabs ();
+extern void init_stmt ();
 extern void init_reg_sets ();
 extern void dump_flow_info ();
 extern void dump_local_alloc ();
@@ -92,8 +96,10 @@ extern void dump_local_alloc ();
 void rest_of_decl_compilation ();
 void error ();
 void error_with_file_and_line ();
+void fancy_abort ();
 void set_target_switch ();
 void print_target_switch_defaults ();
+static char *decl_name ();
 
 /* Bit flags that specify the machine subtype we are compiling for.
    Bits are tested using macros TARGET_... defined in the tm-...h file
@@ -207,6 +213,11 @@ int flag_float_store = 0;
 
 int flag_combine_regs = 0;
 
+/* Nonzero for -fcse-follow-jumps:
+   have cse follow jumps to do a more extensive job.  */
+
+int flag_cse_follow_jumps;
+
 /* Nonzero enables strength-reduction in loop.c.  */
 
 int flag_strength_reduce = 0;
@@ -239,9 +250,8 @@ int flag_volatile;
 
 int flag_syntax_only = 0;
 
-/* Nonzero means do stupid register allocation.  -noreg.
-   This and `optimize' are controlled by different switches in cc1/c++,
-   but normally cc controls them both with the -O switch.  */
+/* Nonzero means do stupid register allocation.
+   Currently, this is 1 if `optimize' is 0.  */
 
 int obey_regdecls = 0;
 
@@ -270,18 +280,34 @@ int warn_shadow;
 
 int warn_switch;
 
+/* Warn if a type conversion is done that might have confusing results.  */
+
+int warn_conversion;
+
 /* Nonzero means warn about any identifiers that match in the first N
    characters.  The value N is in `id_clash_len'.  */
 
 int warn_id_clash;
 int id_clash_len;
 
+/* Nonzero means warn if inline function is too large.  */
+
+int warn_inline;
+
 /* Number of error messages and warning messages so far.  */
 
 int errorcount = 0;
 int warningcount = 0;
 int sorrycount = 0;
 
+/* Pointer to function to compute the name to use to print a declaration.  */
+
+char *(*decl_printable_name) ();
+
+/* Pointer to function to compute rtl for a language-specific tree code.  */
+
+struct rtx_def *(*lang_expand_expr) ();
+
 /* Name of program invoked, sans directories.  */
 
 char *progname;
@@ -318,21 +344,30 @@ int flag_shared_data;
 
 int flag_delayed_branch;
 
-/* Nonzero means schedule instructions within basic blocks.  */
-
-int flag_schedule_insns = 0;
-
 /* Copy of arguments to main.  */
 int save_argc;
 char **save_argv;
 
+/* Nonzero means schedule instructions within basic blocks.  */
+
+int flag_schedule_insns = 0;
+
 /* Name for output file of assembly code, specified with -o.  */
 
 char *asm_file_name;
 
+#if 0
 /* Name for output file of GDB symbol segment, specified with -symout.  */
 
 char *sym_file_name;
+#endif /* 0 */
+
+#ifdef G_SWITCH
+/* Value of the -G nn switch if the machine needs it.  */
+
+int g_switch_value = 0;
+int g_switch_used = 0;
+#endif
 
 /* Table of language-independent -f options.
    STRING is the option name.  VARIABLE is the address of the variable.
@@ -346,6 +381,7 @@ static struct { char *string; int *variable; int on_value;} f_options[] =
   {"volatile", &flag_volatile, 1},
   {"defer-pop", &flag_defer_pop, 1},
   {"omit-frame-pointer", &flag_omit_frame_pointer, 1},
+  {"cse-follow-jumps", &flag_cse_follow_jumps, 1},
   {"strength-reduce", &flag_strength_reduce, 1},
   {"writable-strings", &flag_writable_strings, 1},
   {"peephole", &flag_no_peephole, 0},
@@ -391,9 +427,14 @@ int combine_time;
 int local_alloc_time;
 int global_alloc_time;
 int dbr_sched_time;
+int shorten_branch_time;
 int final_time;
 int symout_time;
 int dump_time;
+
+#ifdef FIELD_XREF
+  int field_xref = 0;		/* cross referencing flag	*/
+#endif
 
 /* Return time used so far, in microseconds.  */
 
@@ -505,6 +546,15 @@ fatal_insn_not_found (insn)
   debug_rtx (insn);
   abort ();
 }
+
+/* This is the default decl_printable_name function.  */
+
+static char *
+decl_name (decl)
+     tree decl;
+{
+  return IDENTIFIER_POINTER (DECL_NAME (decl));
+}
 
 static int need_error_newline;
 
@@ -528,7 +578,7 @@ announce_function (decl)
       if (rtl_dump_and_exit)
 	fprintf (stderr, "%s ", IDENTIFIER_POINTER (DECL_NAME (decl)));
       else
-	fprintf (stderr, " %s", lang_printable_name (decl));
+	fprintf (stderr, " %s", (*decl_printable_name) (decl));
       fflush (stderr);
       need_error_newline = 1;
       last_error_function = current_function_decl;
@@ -559,10 +609,10 @@ report_error_function (file)
 	fprintf (stderr, "At top level:\n");
       else if (TREE_CODE (TREE_TYPE (current_function_decl)) == METHOD_TYPE)
 	fprintf (stderr, "In method %s:\n",
-		 lang_printable_name (current_function_decl));
+		 (*decl_printable_name) (current_function_decl));
       else
 	fprintf (stderr, "In function %s:\n",
-		 lang_printable_name (current_function_decl));
+		 (*decl_printable_name) (current_function_decl));
 
       last_error_function = current_function_decl;
     }
@@ -637,7 +687,10 @@ error_with_decl (decl, s, v)
     fprintf (stderr, "%s:%d: ",
 	     DECL_SOURCE_FILE (decl), DECL_SOURCE_LINE (decl));
 
-  fprintf (stderr, s, lang_printable_name (decl), v);
+  if (DECL_NAME (decl))
+    fprintf (stderr, s, (*decl_printable_name) (decl), v);
+  else
+    fprintf (stderr, s, "((anonymous))", v);
   fprintf (stderr, "\n");
 }
 
@@ -738,7 +791,10 @@ warning_with_decl (decl, s, v)
 	     DECL_SOURCE_FILE (decl), DECL_SOURCE_LINE (decl));
 
   fprintf (stderr, "warning: ");
-  fprintf (stderr, s, lang_printable_name (decl), v);
+  if (DECL_NAME (decl))
+    fprintf (stderr, s, (*decl_printable_name) (decl), v);
+  else
+    fprintf (stderr, s, "((anonymous))", v);
   fprintf (stderr, "\n");
 }
 
@@ -830,10 +886,12 @@ exact_log2 (x)
      register unsigned int x;
 {
   register int log = 0;
-  for (log = 0; log < HOST_BITS_PER_INT; log++)
-    if (x == (1 << log))
-      return log;
-  return -1;
+  /* Test for 0 or a power of 2.  */
+  if (x == 0 || x != (x & -x))
+    return -1;
+  while ((x >>= 1) != 0)
+    log++;
+  return log;
 }
 
 /* Given X, an unsigned number, return the largest int Y such that 2**Y <= X.
@@ -843,11 +901,11 @@ int
 floor_log2 (x)
      register unsigned int x;
 {
-  register int log = 0;
-  for (log = 0; log < HOST_BITS_PER_INT; log++)
-    if ((x & ((-1) << log)) == 0)
-      return log - 1;
-  return HOST_BITS_PER_INT - 1;
+  register int log = -1;
+  while (x != 0)
+    log++,
+    x >>= 1;
+  return log;
 }
 
 int float_handled;
@@ -873,7 +931,6 @@ float_signal ()
 {
   if (float_handled == 0)
     abort ();
-  warning ("floating overflow in constant folding");
   float_handled = 0;
   longjmp (float_handler, 1);
 }
@@ -914,6 +971,7 @@ compile_file (name)
   local_alloc_time = 0;
   global_alloc_time = 0;
   dbr_sched_time = 0;
+  shorten_branch_time = 0;
   final_time = 0;
   symout_time = 0;
   dump_time = 0;
@@ -931,7 +989,7 @@ compile_file (name)
     pfatal_with_name (name);
 
   /* Initialize data in various passes,  
-     checking if this is an unexec'd program - bryan@kewill.uucp */
+     checking if this is an unexec'd program - bryan@kewill.uucp.  */
 
   if (!been_here_before)
     {
@@ -946,11 +1004,15 @@ compile_file (name)
   else
     {
       extern void reinit_lex_for_unexec ();
+      main_input_filename = 0;
       reinit_lex_for_unexec ();
       reinit_search_statistics ();
-      main_input_filename = 0;
     }
 
+#ifdef FIELD_XREF
+  if (field_xref) FIELD_xref_begin(input_filename);
+#endif
+
   /* If rtl dump desired, open the output file.  */
   if (rtl_dump)
     {
@@ -1128,11 +1190,17 @@ compile_file (name)
       prev_file = fopen (previous_asm_file_name, "r");
       if (prev_file == 0)
 	pfatal_with_name (previous_asm_file_name);
+#ifndef VMS
+#ifndef MASSCOMP
       if (_bufsiz(asm_out_file) != _bufsiz(prev_file))
+#endif
+#endif
 	{
 	  while ((c = getc (prev_file)) != EOF)
 	    putc (c, asm_out_file);
 	}
+#ifndef VMS
+#ifndef MASSCOMP
       else
 	{
 	  /* Copy quickly.  */
@@ -1150,9 +1218,12 @@ compile_file (name)
 	    }
 	  while (size == _bufsiz(asm_out_file));
 	}
+#endif
+#endif
       fclose (prev_file);
     }
 
+#ifdef 0  /* This mode is discontinued.  */
   /* If GDB symbol table desired, open the GDB symbol output file.  */
   if (write_symbols == GDB_DEBUG)
     {
@@ -1170,18 +1241,21 @@ compile_file (name)
 	sym_file_name = dumpname;
       symout_init (sym_file_name, asm_out_file, main_input_filename);
     }
+#endif /* 0 */
 
   /* If dbx symbol table desired, initialize writing it
      and output the predefined types.  */
 #ifdef DBX_DEBUGGING_INFO
   if (write_symbols == DBX_DEBUG)
-    if (!been_here_before) 
-      dbxout_init (asm_out_file, main_input_filename);
-    else
-      {
-	extern void re_init_dbxout_for_unexec ();
-	re_init_dbxout_for_unexec (asm_out_file, main_input_filename);
-      }
+    {
+      if (!been_here_before) 
+	dbxout_init (asm_out_file, main_input_filename);
+      else
+	{
+	  extern void re_init_dbxout_for_unexec ();
+	  re_init_dbxout_for_unexec (asm_out_file, main_input_filename);
+	}
+    }
 #endif
 #ifdef SDB_DEBUGGING_INFO
   if (write_symbols == SDB_DEBUG)
@@ -1191,9 +1265,7 @@ compile_file (name)
   /* Initialize yet another pass.  */
 
   if (!been_here_before)
-    {
-      init_final (main_input_filename);
-    }
+    init_final (main_input_filename);
 
   start_time = gettime ();
 
@@ -1203,8 +1275,13 @@ compile_file (name)
   if (setjmp (toplevel) == 0)
     {
       if (yyparse () != 0)
-	if (errorcount == 0)
-	  fprintf (stderr, "Errors detected in input file (your bison.simple is out of date)");
+	{
+	  if (errorcount == 0)
+	    {
+	      fprintf (stderr, "Errors detected in input file (your bison.simple is out of date)");
+	      errorcount = 1;
+	    }
+	}
 
       /* Compilation is now finished except for writing
 	 what's left of the symbol table output.  */
@@ -1247,7 +1324,8 @@ compile_file (name)
 	       to take their address for some reason.  */
 	    if (! TREE_READONLY (decl)
 		|| TREE_PUBLIC (decl)
-		|| TREE_ADDRESSABLE (decl))
+		|| TREE_ADDRESSABLE (decl)
+		|| TREE_ADDRESSABLE (DECL_NAME (decl)))
 	      rest_of_decl_compilation (decl, 0, 1, 1);
 	    /* Otherwise maybe mention them just for the debugger.  */
 #ifdef DBX_DEBUGGING_INFO
@@ -1262,18 +1340,19 @@ compile_file (name)
 	if (TREE_CODE (decl) == FUNCTION_DECL
 	    && ! TREE_ASM_WRITTEN (decl)
 	    && DECL_INITIAL (decl) != 0
-	    && TREE_ADDRESSABLE (decl)
+	    && (TREE_ADDRESSABLE (decl) || TREE_ADDRESSABLE (DECL_NAME (decl)))
 	    && ! TREE_EXTERNAL (decl))
 	  output_inline_function (decl);
 
-	/* Warn about any function declared static but not defined.  */
-	if (warn_unused
-	    && TREE_CODE (decl) == FUNCTION_DECL
+	/* Warn about any function or variable
+	   declared static but not defined.  */
+	if ((warn_unused || TREE_USED (decl) || TREE_USED (DECL_NAME (decl)))
+/*	    && TREE_CODE (decl) == FUNCTION_DECL  */
 	    && DECL_INITIAL (decl) == 0
 	    && TREE_EXTERNAL (decl)
 	    && ! TREE_PUBLIC (decl))
-	  warning_with_decl (decl, "`%s' declared but never defined");
-	/* Warn about statics fns or vars defined but not used,
+	  warning_with_decl (decl, "`%s' declared `static' but never defined");
+	/* Warn about static fns or vars defined but not used,
 	   but not about inline functions
 	   since unused inline statics is normal practice.  */
 	if (warn_unused
@@ -1282,7 +1361,11 @@ compile_file (name)
 	    && ! TREE_EXTERNAL (decl)
 	    && ! TREE_PUBLIC (decl)
 	    && ! TREE_USED (decl)
-	    && ! TREE_INLINE (decl))
+	    && ! TREE_INLINE (decl)
+	    /* The TREE_USED bit for file-scope decls
+	       is kept in the identifier, to handle multiple
+	       external decls in different scopes.  */
+	    && ! TREE_USED (DECL_NAME (decl)))
 	  warning_with_decl (decl, "`%s' defined but not used");
       }
   }
@@ -1306,6 +1389,7 @@ compile_file (name)
 	     });
 #endif
 
+#if 0  /* This mode is discontinued */
   /* Do gdb symbols */
   if (write_symbols == GDB_DEBUG)
     TIMEVAR (symout_time,
@@ -1316,6 +1400,7 @@ compile_file (name)
 	       symout_top_blocks (globals, gettags ());
 	       symout_finish (name, statbuf.st_ctime);
 	     });
+#endif /* 0 */
 
   /* Output some stuff at end of file if nec.  */
 
@@ -1325,7 +1410,6 @@ compile_file (name)
   ASM_FILE_END (asm_out_file);
 #endif
 
-
  after_finish_compilation:
 
   /* Close the dump files.  */
@@ -1385,6 +1469,7 @@ compile_file (name)
       print_time ("combine", combine_time);
       print_time ("local-alloc", local_alloc_time);
       print_time ("global-alloc", global_alloc_time);
+      print_time ("shorten-branch", shorten_branch_time);
       print_time ("dbranch", dbr_sched_time);
       print_time ("final", final_time);
       print_time ("varconst", varconst_time);
@@ -1450,6 +1535,7 @@ rest_of_decl_compilation (decl, asmspec, top_level, at_end)
 
   if (top_level)
     {
+#if 0  /* This mode is discontinued */
       if (write_symbols == GDB_DEBUG)
 	{
 	  TIMEVAR (symout_time,
@@ -1461,6 +1547,7 @@ rest_of_decl_compilation (decl, asmspec, top_level, at_end)
 		   });
 	}
       else
+#endif /* 0 */
 	/* Clean out the temporary type list, since the types will go away.  */
 	get_temporary_types ();
     }
@@ -1485,22 +1572,21 @@ rest_of_compilation (decl)
 
   if (DECL_SAVED_INSNS (decl) == 0)
     {
+      int specd = TREE_INLINE (decl);
+      char *lose;
 
-      /* If requested, consider whether to make this function inline.  */
-      if (flag_inline_functions || TREE_INLINE (decl))
+      if (specd || flag_inline_functions)
 	{
-	  TIMEVAR (integration_time,
-		   {
-		     int specd = TREE_INLINE (decl);
-		     char *lose = function_cannot_inline_p (decl);
-		     if (lose != 0 && specd)
-		       warning_with_decl (decl, lose);
-		     if (lose == 0)
-		       save_for_inline (decl);
-		     else
-		       TREE_INLINE (decl) = 0;
-		   });
+	  lose = function_cannot_inline_p (decl);
+	  if (lose)
+	    {
+	      if (warn_inline && specd)
+		warning_with_decl (decl, lose);
+	      TREE_INLINE (decl) = 0;
+	    }
 	}
+      else
+	lose = "compiler error";
 
       insns = get_insns ();
 
@@ -1511,7 +1597,7 @@ rest_of_compilation (decl)
 		 {
 		   fprintf (rtl_dump_file, "\n;; Function %s\n\n",
 			    IDENTIFIER_POINTER (DECL_NAME (decl)));
-		   if (DECL_SAVED_INSNS (decl))
+		   if (lose == 0)
 		     fprintf (rtl_dump_file, ";; (integrable)\n\n");
 		   print_rtl (rtl_dump_file, insns);
 		   fflush (rtl_dump_file);
@@ -1522,11 +1608,19 @@ rest_of_compilation (decl)
 	 finish_compilation will call rest_of_compilation again
 	 for those functions that need to be output.  */
 
-      if (((! TREE_PUBLIC (decl) && ! TREE_ADDRESSABLE (decl)
-	    && ! flag_keep_inline_functions)
-	   || TREE_EXTERNAL (decl))
-	  && TREE_INLINE (decl))
-	goto exit_rest_of_compilation;
+      if (TREE_INLINE (decl)
+	  && ((! TREE_PUBLIC (decl) && ! TREE_ADDRESSABLE (decl)
+	       && ! flag_keep_inline_functions)
+	      || TREE_EXTERNAL (decl)))
+	{
+	  save_for_outline (decl);
+	  goto exit_rest_of_compilation;
+	}
+
+      /* If we have to compile the function now, save its rtl
+	 so that its compilation will not affect what others get.  */
+      if (TREE_INLINE (decl))
+	TIMEVAR (integration_time, save_for_inline (decl));
     }
 
   TREE_ASM_WRITTEN (decl) = 1;
@@ -1573,11 +1667,18 @@ rest_of_compilation (decl)
      and some code may now be unreachable, so do
      jump optimization again.  */
 
+  if (cse_dump)
+    TIMEVAR (dump_time,
+	     {
+	       fprintf (cse_dump_file, "\n;; Function %s\n\n",
+			IDENTIFIER_POINTER (DECL_NAME (decl)));
+	     });
+
   if (optimize)
     {
       TIMEVAR (cse_time, reg_scan (insns, max_reg_num (), 0));
 
-      TIMEVAR (cse_time, tem = cse_main (insns, max_reg_num ()));
+      TIMEVAR (cse_time, tem = cse_main (insns, max_reg_num (), cse_dump_file));
 
       if (tem)
 	TIMEVAR (jump_time, jump_optimize (insns, 0, 0));
@@ -1588,8 +1689,6 @@ rest_of_compilation (decl)
   if (cse_dump)
     TIMEVAR (dump_time,
 	     {
-	       fprintf (cse_dump_file, "\n;; Function %s\n\n",
-			IDENTIFIER_POINTER (DECL_NAME (decl)));
 	       print_rtl (cse_dump_file, insns);
 	       fflush (cse_dump_file);
 	     });
@@ -1779,6 +1878,15 @@ rest_of_compilation (decl)
     }
 #endif
 
+#if 0
+  if (optimize)
+    /* Shorten branches.  */
+    TIMEVAR (shorten_branch_time,
+	     {
+	       shorten_branches (get_insns ());
+	     });
+#endif
+
   /* Now turn the rtl into assembler code.  */
 
   TIMEVAR (final_time,
@@ -1793,6 +1901,7 @@ rest_of_compilation (decl)
 	     fflush (asm_out_file);
 	   });
 
+#if 0  /* This mode is discontinued */
   /* Write GDB symbols if requested */
 
   if (write_symbols == GDB_DEBUG)
@@ -1809,6 +1918,7 @@ rest_of_compilation (decl)
 	       });
     }
   else
+#endif /* 0 */
     get_temporary_types ();
 
   /* Write DBX symbols if requested */
@@ -1834,10 +1944,53 @@ rest_of_compilation (decl)
   parse_time -= gettime () - start_time;
 }
 
+#ifdef sun
+#include <sys/mman.h>
+/* Hacky routine to read back entire data segment saved earlier. 
+   Michael Tiemann (tiemann@cygnus.com).  */
+
+void undump_data (path, name)
+     char *path, *name;
+{
+  extern char *rindex ();
+  int file;
+  long file_size;
+  extern int my_edata;
+  char *data;
+
+  if ((file = open (name, O_RDONLY)) == -1)
+    {
+      char buf[256], *p;
+      strcpy (buf, path);
+      p = rindex (buf, '/');
+      if (p)
+	{
+	  strcpy (p+1, name);
+	  file = open (buf, O_RDONLY);
+	}
+    }
+  if (file == -1)
+    {
+      perror(name);
+      exit (-1);
+    }
+
+  file_size = lseek (file, 0L, SEEK_END);
+  brk ((char*)(&my_edata) + file_size);
+
+  data = (char *)((int)&my_edata&~(getpagesize()-1));
+  if ((int)mmap (data, file_size, PROT_READ|PROT_WRITE,
+		 MAP_PRIVATE|MAP_FIXED, file, 0) < 0)
+    {
+      perror (name);
+      exit (-1);
+    }
+}
+#else
 /* Hacky routine to read back entire data segment saved earlier. 
    Bryan Boreham (bryan@kewill.uucp)    */
 
-void undump_data (name)
+void undump_data (path, name)
      char *name;
 {
   int file;
@@ -1855,10 +2008,11 @@ void undump_data (name)
       }
   file_size = lseek(file, 0L, SEEK_END);
   lseek(file, 0L, SEEK_SET);
-  brk((char*)(&my_edata) + file_size);
+  brk((char*)((int)&my_edata&~(getpagesize()-1)) + file_size);
   if (read (file, &my_edata, file_size) != file_size)
     perror(name);
 }
+#endif
 
 /* Entry point of c++.  Decode command args, then call compile_file.
    Exit code is 35 if can't open files, 34 if fatal error,
@@ -1904,6 +2058,9 @@ main (argc, argv, envp)
 
   signal (SIGPIPE, pipe_closed);
 
+  decl_printable_name = decl_name;
+  lang_expand_expr = fancy_abort;
+
   /* Initialize whether `char' is signed.  */
   flag_signed_char = DEFAULT_SIGNED_CHAR;
 #ifdef DEFAULT_SHORT_ENUMS
@@ -1921,225 +2078,244 @@ main (argc, argv, envp)
   set_target_switch ("");
 
   for (i = 1; i < argc; i++)
-    if (argv[i][0] == '-' && argv[i][1] != 0)
-      {
-	register char *str = argv[i] + 1;
-	if (str[0] == 'Y')
-	  str++;
+    {
+      if (argv[i][0] == '-' && argv[i][1] != 0)
+	{
+	  register char *str = argv[i] + 1;
+	  if (str[0] == 'Y')
+	    str++;
 
-	if (str[0] == 'm')
-	  set_target_switch (&str[1]);
-	else if (!strcmp (str, "dumpbase"))
-	  {
-	    dump_base_name = argv[++i];
-	  }
-	else if (str[0] == 'd')
-	  {
-	    register char *p = &str[1];
-	    while (*p)
-	      switch (*p++)
-		{
-		case 'c':
-		  combine_dump = 1;
-		  break;
-		case 'd':
-		  dbr_sched_dump = 1;
-		  break;
-		case 'f':
-		  flow_dump = 1;
-		  break;
-		case 'g':
-		  global_reg_dump = 1;
-		  break;
-		case 'j':
-		  jump_opt_dump = 1;
-		  break;
-		case 'J':
-		  jump2_opt_dump = 1;
-		  break;
-		case 'l':
-		  local_reg_dump = 1;
-		  break;
-		case 'L':
-		  loop_dump = 1;
-		  break;
-		case 'm':
+	  if (str[0] == 'm')
+	    set_target_switch (&str[1]);
+	  else if (!strcmp (str, "dumpbase"))
+	    {
+	      dump_base_name = argv[++i];
+	    }
+	  else if (str[0] == 'd')
+	    {
+	      register char *p = &str[1];
+	      while (*p)
+		switch (*p++)
 		  {
-		    extern int flag_detailed_statistics;
+		  case 'c':
+		    combine_dump = 1;
+		    break;
+		  case 'd':
+		    dbr_sched_dump = 1;
+		    break;
+		  case 'f':
+		    flow_dump = 1;
+		    break;
+		  case 'g':
+		    global_reg_dump = 1;
+		    break;
+		  case 'j':
+		    jump_opt_dump = 1;
+		    break;
+		  case 'J':
+		    jump2_opt_dump = 1;
+		    break;
+		  case 'l':
+		    local_reg_dump = 1;
+		    break;
+		  case 'L':
+		    loop_dump = 1;
+		    break;
+		  case 'm':
 		    print_mem_flag = 1;
-		    flag_detailed_statistics = 1;
+		    break;
+		  case 'r':
+		    rtl_dump = 1;
+		    break;
+		  case 's':
+		    cse_dump = 1;
+		    break;
+		  case 'y':
+		    yydebug = 1;
+		    break;
+		  case 'x':
+		    rtl_dump_and_exit = 1;
+		    break;
 		  }
-		  break;
-		case 'r':
-		  rtl_dump = 1;
-		  break;
-		case 's':
-		  cse_dump = 1;
-		  break;
-		case 'y':
-		  yydebug = 1;
-		  break;
-		case 'x':
-		  rtl_dump_and_exit = 1;
-		  break;
+	    }
+	  else if (str[0] == 'f')
+	    {
+	      int j;
+	      register char *p = &str[1];
+	      int found = 0;
+
+	      /* Some kind of -f option.
+		 P's value is the option sans `-f'.
+		 Search for it in the table of options.  */
+
+	      for (j = 0;
+		   !found && j < sizeof (f_options) / sizeof (f_options[0]);
+		   j++)
+		{
+		  if (!strcmp (p, f_options[j].string))
+		    {
+		      *f_options[j].variable = f_options[j].on_value;
+		      /* A goto here would be cleaner,
+			 but breaks the vax pcc.  */
+		      found = 1;
+		    }
+		  if (p[0] == 'n' && p[1] == 'o' && p[2] == '-'
+		      && ! strcmp (p+3, f_options[j].string))
+		    {
+		      *f_options[j].variable = ! f_options[j].on_value;
+		      found = 1;
+		    }
 		}
-	  }
-	else if (str[0] == 'f')
-	  {
-	    int j;
-	    register char *p = &str[1];
-	    int found = 0;
 
-	    /* Some kind of -f option.
-	       P's value is the option sans `-f'.
-	       Search for it in the table of options.  */
-
-	    for (j = 0;
-		 !found && j < sizeof (f_options) / sizeof (f_options[0]);
-		 j++)
-	      {
-		if (!strcmp (p, f_options[j].string))
-		  {
-		    *f_options[j].variable = f_options[j].on_value;
-		    /* A goto here would be cleaner,
-		       but breaks the vax pcc.  */
-		    found = 1;
-		  }
-		if (p[0] == 'n' && p[1] == 'o' && p[2] == '-'
-		    && ! strcmp (p+3, f_options[j].string))
-		  {
-		    *f_options[j].variable = ! f_options[j].on_value;
-		    found = 1;
-		  }
-	      }
-
-	    if (found)
-	      ;
-	    else if (!strncmp (p, "undump-", 7))
-	      undump_data (&p[7]);
-	    else if (!strncmp (p, "fixed-", 6))
-	      fix_register (&p[6], 1, 1);
-	    else if (!strncmp (p, "call-used-", 10))
-	      fix_register (&p[10], 0, 1);
-	    else if (!strncmp (p, "call-saved-", 11))
-	      fix_register (&p[11], 0, 0);
-	    else if (! lang_decode_option (argv[i]))
-	      error ("Invalid option `%s'", argv[i]);
-	  }
-	else if (!strcmp (str, "noreg"))
-	  ;
-	else if (!strcmp (str, "opt"))
-	  optimize = 1, obey_regdecls = 0;
-	else if (!strcmp (str, "O"))
-	  optimize = 1, obey_regdecls = 0;
-	else if (!strcmp (str, "pedantic"))
-	  pedantic = 1;
-	else if (lang_decode_option (argv[i]))
-	  ;
-	else if (!strcmp (str, "quiet"))
-	  quiet_flag = 1;
-	else if (!strcmp (str, "version"))
-	  {
-	    extern char *version_string, *language_string;
-	    fprintf (stderr, "%s version %s", language_string, version_string);
+	      if (found)
+		;
+	      else if (!strncmp (p, "undump-", 7))
+		undump_data (argv[0], &p[7]);
+	      else if (!strncmp (p, "fixed-", 6))
+		fix_register (&p[6], 1, 1);
+	      else if (!strncmp (p, "call-used-", 10))
+		fix_register (&p[10], 0, 1);
+	      else if (!strncmp (p, "call-saved-", 11))
+		fix_register (&p[11], 0, 0);
+	      else if (! lang_decode_option (argv[i]))
+		error ("Invalid option `%s'", argv[i]);
+	    }
+	  else if (!strcmp (str, "noreg"))
+	    ;
+	  else if (!strcmp (str, "opt"))
+	    optimize = 1, obey_regdecls = 0;
+	  else if (!strcmp (str, "O"))
+	    optimize = 1, obey_regdecls = 0;
+	  else if (!strcmp (str, "pedantic"))
+	    pedantic = 1;
+	  else if (lang_decode_option (argv[i]))
+	    ;
+	  else if (!strcmp (str, "quiet"))
+	    quiet_flag = 1;
+	  else if (!strcmp (str, "version"))
+	    {
+	      extern char *version_string, *language_string;
+	      fprintf (stderr, "%s version %s", language_string, version_string);
 #ifdef TARGET_VERSION
-	    TARGET_VERSION;
+	      TARGET_VERSION;
 #endif
 #ifdef __GNUC__
 #ifndef __VERSION__
 #define __VERSION__ "[unknown]"
 #endif
-	    fprintf (stderr, " compiled by GNU C version %s.\n", __VERSION__);
+	      fprintf (stderr, " compiled by GNU C version %s.\n", __VERSION__);
 #else
-	    fprintf (stderr, " compiled by CC.\n");
+	      fprintf (stderr, " compiled by CC.\n");
 #endif
-	    print_target_switch_defaults ();
-	  }
-	else if (!strcmp (str, "w"))
-	  inhibit_warnings = 1;
-	else if (!strcmp (str, "W"))
-	  extra_warnings = 1;
-	else if (!strcmp (str, "Wunused"))
-	  warn_unused = 1;
-	else if (!strcmp (str, "Wshadow"))
-	  warn_shadow = 1;
-	else if (!strcmp (str, "Wswitch"))
-	  warn_switch = 1;
-	else if (!strncmp (str, "Wid-clash-", 10))
-	  {
-	    char *endp = str + 10;
+	      print_target_switch_defaults ();
+	    }
+	  else if (!strcmp (str, "w"))
+	    inhibit_warnings = 1;
+	  else if (!strcmp (str, "W"))
+	    extra_warnings = 1;
+	  else if (!strcmp (str, "Wunused"))
+	    warn_unused = 1;
+	  else if (!strcmp (str, "Wshadow"))
+	    warn_shadow = 1;
+	  else if (!strcmp (str, "Wswitch"))
+	    warn_switch = 1;
+	  else if (!strcmp (str, "Wconversion"))
+	    warn_conversion = 1;
+	  else if (!strncmp (str, "Wid-clash-", 10))
+	    {
+	      char *endp = str + 10;
 
-	    while (*endp)
-	      {
-		if (*endp >= '0' && *endp <= '9')
-		  endp++;
-		else
-		  error ("Invalid option `%s'", argv[i]);
-	      }
-	    warn_id_clash = 1;
-	    id_clash_len = atoi (str + 10);
-	  }
-	else if (!strcmp (str, "p"))
-	  profile_flag = 1;
-	else if (!strcmp (str, "a"))
-	  {
+	      while (*endp)
+		{
+		  if (*endp >= '0' && *endp <= '9')
+		    endp++;
+		  else
+		    error ("Invalid option `%s'", argv[i]);
+		}
+	      warn_id_clash = 1;
+	      id_clash_len = atoi (str + 10);
+	    }
+	  else if (!strcmp (str, "Winline"))
+	    warn_inline = 1;
+	  else if (!strcmp (str, "p"))
+	    profile_flag = 1;
+	  else if (!strcmp (str, "a"))
+	    {
 #if !defined (BLOCK_PROFILER) || !defined (FUNCTION_BLOCK_PROFILER)
-	    warning ("`-a' option (basic block profile) not supported");
+	      warning ("`-a' option (basic block profile) not supported");
 #else
-	    profile_block_flag = 1;
+	      profile_block_flag = 1;
+#endif
+	    }
+#ifdef 0
+	  else if (!strcmp (str, "gg"))
+	    write_symbols = GDB_DEBUG;
 #endif
-	  }
-	else if (!strcmp (str, "gg"))
-	  write_symbols = GDB_DEBUG;
 #ifdef DBX_DEBUGGING_INFO
-	else if (!strcmp (str, "g0"))
-	  write_symbols = DBX_DEBUG;
-	else if (!strcmp (str, "G0"))
-	  write_symbols = DBX_DEBUG;
-	else if (!strcmp (str, "g"))
-	  {
+	  else if (!strcmp (str, "g0"))
 	    write_symbols = DBX_DEBUG;
-	    use_gdb_dbx_extensions = 1;
-	  }
-	else if (!strcmp (str, "G"))
-	  {
+	  else if (!strcmp (str, "G0"))
 	    write_symbols = DBX_DEBUG;
-	    use_gdb_dbx_extensions = 1;
-	  }
-#endif
 #ifdef SDB_DEBUGGING_INFO
-	else if (!strcmp (str, "g"))
-	  write_symbols = SDB_DEBUG;
-	else if (!strcmp (str, "G"))
-	  write_symbols = SDB_DEBUG;
-	else if (!strcmp (str, "g0"))
-	  write_symbols = SDB_DEBUG;
-	else if (!strcmp (str, "G0"))
-	  write_symbols = SDB_DEBUG;
+	  else if (!strcmp (str, "gsdb"))
+	    write_symbols = SDB_DEBUG;
 #endif
-	else if (!strcmp (str, "symout"))
-	  {
-	    if (write_symbols == NO_DEBUG)
-	      write_symbols = GDB_DEBUG;
-	    sym_file_name = argv[++i];
-	  }
-	else if (!strcmp (str, "o"))
-	  {
-	    asm_file_name = argv[++i];
-	  }
-	else
-	  error ("Invalid option `%s'", argv[i]);
-      }
-  else if (argv[i][0] == '+')
-    {
-      if (lang_decode_option (argv[i]))
-	;
+	  else if (!strcmp (str, "g"))
+	    {
+	      write_symbols = DBX_DEBUG;
+	      use_gdb_dbx_extensions = 1;
+	    }
+	  else if (!strcmp (str, "G"))
+	    {
+	      write_symbols = DBX_DEBUG;
+	      use_gdb_dbx_extensions = 1;
+	    }
+#endif				/* DBX_DEBUGGING_INFO */
+#ifdef SDB_DEBUGGING_INFO
+	  else if (!strcmp (str, "g"))
+	    write_symbols = SDB_DEBUG;
+	  else if (!strcmp (str, "G"))
+	    write_symbols = SDB_DEBUG;
+	  else if (!strcmp (str, "g0"))
+	    write_symbols = SDB_DEBUG;
+	  else if (!strcmp (str, "G0"))
+	    write_symbols = SDB_DEBUG;
+	  else if (!strcmp (str, "gsdb"))
+	    write_symbols = SDB_DEBUG;
+#endif /* SDB_DEBUGGING_INFO */
+#if defined(ASM_OUTPUT_SOURCE_FILENAME) && defined(ASM_OUTPUT_SOURCE_LINE)
+	  else if (!strcmp (str, "gline"))
+	    write_symbols = LINENO_DEBUG;
+#endif
+#if 0
+	  else if (!strcmp (str, "symout"))
+	    {
+	      if (write_symbols == NO_DEBUG)
+		write_symbols = GDB_DEBUG;
+	      sym_file_name = argv[++i];
+	    }
+#endif				/* 0 */
+	  else if (!strcmp (str, "o"))
+	    {
+	      asm_file_name = argv[++i];
+	    }
+#ifdef FIELD_XREF
+	  else if (!strcmp(str,"xref"))
+	    field_xref = 1;
+#endif
+	  else
+	    error ("Invalid option `%s'", argv[i]);
+	}
+      else if (argv[i][0] == '+')
+	{
+	  if (lang_decode_option (argv[i]))
+	    ;
+	  else
+	    error ("Invalid option `%s'", argv[i]);
+	}
       else
-	error ("Invalid option `%s'", argv[i]);
+	filename = argv[i];
     }
-  else
-    filename = argv[i];
-
 #ifdef OVERRIDE_OPTIONS
   /* Some machines may reject certain combinations of options.  */
   OVERRIDE_OPTIONS;
@@ -2161,13 +2337,15 @@ main (argc, argv, envp)
 	       (int) lim - (int) &environ);
       fflush (stderr);
 
-      dump_time_statistics ();
-
       system ("ps v");
     }
 #endif /* not VMS */
 #endif /* not USG */
 
+#ifdef FIELD_XREF
+  FIELD_xref_end(errorcount+sorrycount);
+#endif
+
   if (errorcount)
     exit (FATAL_EXIT_CODE);
   if (sorrycount)
@@ -2194,6 +2372,8 @@ set_target_switch (name)
      char *name;
 {
   register int j;
+  int valid = 0;
+
   for (j = 0; j < sizeof target_switches / sizeof target_switches[0]; j++)
     if (!strcmp (target_switches[j].name, name))
       {
@@ -2201,9 +2381,11 @@ set_target_switch (name)
 	  target_flags &= ~-target_switches[j].value;
 	else
 	  target_flags |= target_switches[j].value;
-	return;
+	valid = 1;
       }
-  error ("Invalid option `%s'", name);
+
+  if (!valid)
+    error ("Invalid option `%s'", name);
 }
 
 /* Print default target switches for -version.  */
@@ -2213,12 +2395,12 @@ print_target_switch_defaults ()
 {
   register int j;
   register int mask = TARGET_DEFAULT;
+
   fprintf (stderr, "default target switches:");
   for (j = 0; j < sizeof target_switches / sizeof target_switches[0]; j++)
     if (target_switches[j].name[0] != '\0'
 	&& target_switches[j].value > 0
 	&& (target_switches[j].value & mask) == target_switches[j].value)
-
       fprintf (stderr, " -m%s", target_switches[j].name);
 
   fprintf (stderr, "\n");
